from brownie import Contract
from .utils.multicall2 import fetch_multicall


def is_yearn_vault(token):
    vault = Contract(token)
    return hasattr(vault, 'pricePerShare') or hasattr(vault, 'getPricePerFullShare')


def get_price(token, block=None):
    # v1 vaults use getPricePerFullShare scaled to 18 decimals
    # v2 vaults use pricePerShare scaled to underlying token decimals
    vault = Contract(token)
    if hasattr(vault, 'pricePerShare'):
        share_price, underlying, decimals = fetch_multicall(
            [vault, 'pricePerShare'],
            [vault, 'token'],
            [vault, 'decimals'],
            block=block
        )
        return [share_price / 10 ** decimals, underlying]
    if hasattr(vault, 'getPricePerFullShare'):
        share_price, underlying = fetch_multicall(
            [vault, 'getPricePerFullShare'],
            [vault, 'token'],
            block=block
        )
        return [share_price / 1e18, underlying]



from brownie import Contract #, interface
from cachetools.func import ttl_cache
from .utils.multicall2 import fetch_multicall


@ttl_cache(ttl=3600)
def get_markets():
    comptroller = Contract("0x3d9819210A31b4961b30EF54bE2aeD79B9c9Cd3B")
    creamtroller = Contract('0x3d5BC3c8d13dcB8bF317092d84783c2697AE9258')
    ironbankroller = Contract("0xAB1c342C7bf5Ec5F02ADEA1c2270670bCa144CbB")
    
    results = fetch_multicall(
        [comptroller, 'getAllMarkets'],
        [creamtroller, 'getAllMarkets'],
        [ironbankroller, 'getAllMarkets'],
    )
    names = ['compound', 'cream', 'ironbank']
    return dict(zip(names, results))


def is_compound_market(token):
    markets = get_markets()
    return any(token in market for market in markets.values())


def get_price(token, block=None):
    token = Contract(token)
    underlying, exchange_rate, decimals = fetch_multicall(
        [token, 'underlying'],
        [token, 'exchangeRateCurrent'],
        [token, 'decimals'],
        block=block
    )
    exchange_rate /= 1e18
    under_decimals = Contract(underlying).decimals()
    return [exchange_rate * 10 ** (decimals - under_decimals), underlying]



from brownie import Contract
from cachetools.func import ttl_cache

from .utils.cache import memory
from .utils.multicall2 import fetch_multicall
from . import magic


@memory.cache()
def is_balancer_pool(address):
    pool = Contract(address)
    required = {"getCurrentTokens", "getBalance", "totalSupply"}
    if set(pool.__dict__) & required == required:
        return True
    return False


@ttl_cache(ttl=600)
def get_price(token, block=None):
    pool = Contract(token)
    tokens, supply = fetch_multicall([pool, "getCurrentTokens"], [pool, "totalSupply"], block=block)
    supply = supply / 1e18
    balances = fetch_multicall(*[[pool, "getBalance", token] for token in tokens], block=block)
    balances = [balance / 10 ** Contract(token).decimals() for balance, token in zip(balances, tokens)]
    total = sum(balance * magic.get_price(token, block=block) for balance, token in zip(balances, tokens))
    return total / supply



from setuptools import find_packages, setup

setup(
    name='ypricemagic',
    packages=find_packages(),
    version='0.1.2',
    description='Use this tool to extract historical on-chain price data from an archive node. Shoutout to @bantg and @nymmrx for their awesome work on yearn-exporter that made this library possible.',
    author='BobTheBuidler',
    author_email = 'bobthebuidlerdefi@gmail.com',
    license='MIT',
    install_requires=[
        'cachetools>=4.1.1',
        'eth-brownie==1.14.6',
        'joblib>=1.0.1'
    ],
    )


import logging

from cachetools.func import ttl_cache

from . import balancer, chainlink, compound, constants, curve, uniswap, yearn

logger = logging.getLogger(__name__)


class PriceError(Exception):
    pass


@ttl_cache(10000)
def get_price(token, block=None):
    token = str(token)
    logger.debug("unwrapping %s", token)
    price = None

    if token in constants.STABLECOINS:
        logger.debug("stablecoin -> %s", 1)
        return 1

    if token == "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE":
        token = str(constants.weth)

    # we can exit early with known tokens
    if token in chainlink.feeds:
        price = chainlink.get_price(token, block=block)
        logger.debug("chainlink -> %s", price)

    elif yearn.is_yearn_vault(token):
        price = yearn.get_price(token, block=block)
        logger.debug("yearn -> %s", price)

    elif curve.is_curve_lp_token(token):
        price = curve.get_price(token, block=block)
        logger.debug("curve lp -> %s", price)

    elif compound.is_compound_market(token):
        price = compound.get_price(token, block=block)
        logger.debug("compound -> %s", price)

    elif uniswap.is_uniswap_pool(token):
        price = uniswap.lp_price(token, block=block)
        logger.debug("uniswap pool -> %s", price)

    elif balancer.is_balancer_pool(token):
        price = balancer.get_price(token, block=block)
        logger.debug("balancer pool -> %s", price)

    # peel a layer from [multiplier, underlying]
    if isinstance(price, list):
        price, underlying = price
        logger.debug("peel %s %s", price, underlying)
        return price * get_price(underlying, block=block)

    # a few more attempts to fetch a price a token
    if not price:
        price = uniswap.get_price(token, router="sushiswap", block=block)
        logger.debug("sushiswap -> %s", price)
    if not price:
        price = uniswap.get_price(token, router="uniswap", block=block)
        logger.debug("uniswap -> %s", price)
    if not price:
        price = uniswap.get_price_v1(token, block=block)
        logger.debug("uniswap v1 -> %s", price)
    if not price:
        logger.error("failed to get price for %s", token)

    if not price:
        raise PriceError(f'could not fetch price for {token}')

    return price
